/*
 * wxCsound main application
 *
 * Copyright (c) 2003 by John D. Ramsdell. All rights reserved.
 *
 * This software is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// This file contains the main wxCsound application, its frame, and
// some other windows.  The main frame's constructor builds the
// windows.  The application's window operations have been collected
// into a model class, and the frame's methods use the model to
// service events.  In other words, the model is where the action is.
// Most of the model for the Settings window is in the main model's
// Args object.

// For file names, the GUI displays file names using relative paths
// whenever possible, but always stores file names using absolute
// paths in the Args object.  It also uses relative paths when reading
// and writing file names in a Csound Configuration file.

#include <wx/wxprec.h>

#ifndef WX_PRECOMP
#include <wx/wx.h>
#endif

#include <wx/notebook.h>
#include <wx/ffile.h>
#include <wx/filename.h>
#include <wx/msgout.h>
#include <wx/dnd.h>

#include "synthesizer.hpp"
#include "curve.hpp"
#include "table.hpp"
#include "wxcsound.xpm"
#include "main.hpp"
#include "worker.hpp"
#include "canvas.hpp"
#include "transcript.hpp"
#include "args.hpp"
#include "viewer.hpp"

// Initial position in pixels
const int app_x_pos = 50;
const int app_y_pos = 50;

// Events and their IDs

enum
{
  id_new = 1,
  id_open,
  id_open_unified,
  id_save,
  id_save_as,
  id_quit,
  id_run,
  id_stop,
  id_clear,
  id_about,
  id_viewer,
  id_list_opcodes,
  id_usage,
  id_settings,
  id_transcript,
  id_canvas,
  id_messages,
  id_halt,
  id_text,
  id_clear_curves,
  id_draw_curves,
  // For settings
  id_args,
  id_file_name,
  id_orchestra,
  id_score,
  id_output,
  id_clear_orchestra,
  id_clear_score,
  id_clear_output,
  id_no_sound,
  id_rewrite_header,
  id_peak_chunks,
  id_format,
  id_size,
  id_graphics,
};

// Notebook page numbers

enum {
  page_settings,
  page_transcript,
  page_canvas,
  page_messages,
};

// Strings used in the Settings window

const wxChar empty_file_name[] = wxT("<unspecified>");

// Format

const wxChar default_format[] = wxT("Default");
const wxChar aiff_format[] = wxT("AIFF");
const wxChar wave_format[] = wxT("WAV");
const wxChar ircam_format[] = wxT("IRCAM");
const wxChar raw_format[] = wxT("Raw");

// Size
const wxChar default_size[] = wxT("Default");
const wxChar eight_unsigned_size[] = wxT("8-bit unsigned");
const wxChar eight_signed_size[] = wxT("8-bit signed");
const wxChar sixteen_size[] = wxT("16-bit");
const wxChar twenty_four_size[] = wxT("24-bit packed");
const wxChar thirty_two_size[] = wxT("32-bit");
const wxChar float_size[] = wxT("Floating point");

// Graphics
const wxChar full_graphics[] = wxT("Full");
const wxChar postscript_graphics[] = wxT("PostScript");
const wxChar ascii_graphics[] = wxT("ASCII");
const wxChar no_graphics[] = wxT("None");

// File names

const wxChar config_ext[] = wxT("csc");
const wxChar config_wildcard[] =
wxT("CSC files (*.csc)|*.csc|All files (*)|*");

const wxChar csd_ext[] = wxT("csd");
const wxChar csd_wildcard[] =
wxT("CSD files (*.csd)|*.csd|All files (*)|*");

const wxChar orchestra_ext[] = wxT("orc");
const wxChar orchestra_wildcard[] =
wxT("ORC files (*.orc)|*.orc|All files (*)|*");

const wxChar score_ext[] = wxT("sco");
const wxChar score_wildcard[] =
wxT("SCO files (*.sco)|*.sco|All files (*)|*");

const wxChar *format_ext[] =
{
  wxT(""),
  wxT("aiff"),
  wxT("wav"),
  wxT("ircam"),
  wxT("raw"),
};

const wxChar *format_wildcard[] =
{
  wxT("All files (*)|*"),
  wxT("AIFF files (*.aiff)|*.aiff|All files (*)|*"),
  wxT("WAV files (*.wav)|*.wav|All files (*)|*"),
  wxT("IRCAM files (*.ircam)|*.ircam|All files (*)|*"),
  wxT("Raw files (*.raw)|*.raw|All files (*)|*"),
};

// The Settings window

class Settings : public wxPanel
{
public:
  Settings(Args *, wxFileName *, wxWindow *);
  void on_args_update(wxUpdateUIEvent&);
  void on_file_name_update(wxUpdateUIEvent&);
  void on_orchestra(wxCommandEvent&);
  void on_score(wxCommandEvent&);
  void on_output(wxCommandEvent&);
  void on_clear_orchestra(wxCommandEvent&);
  void on_clear_score(wxCommandEvent&);
  void on_clear_output(wxCommandEvent&);
  void on_no_sound(wxCommandEvent&);
  void on_rewrite_header(wxCommandEvent&);
  void on_peak_chunks(wxCommandEvent&);
  void on_format(wxCommandEvent&);
  void on_size(wxCommandEvent&);
  void on_graphics(wxCommandEvent&);
private:
  Args *m_args;
  wxFileName *m_file_name;
  wxStaticText *m_orchestra;
  wxStaticText *m_score;
  wxStaticText *m_output;
  wxCheckBox *m_no_sound;
  wxCheckBox *m_rewrite_header;
  wxCheckBox *m_peak_chunks;
  wxRadioBox *m_format;
  wxRadioBox *m_size;
  wxRadioBox *m_graphics;
  wxStaticText *m_file;
  wxStaticText *m_command;
  void do_args_update();
  static void stack_two(wxPanel *, wxWindow *, wxWindow *);
  DECLARE_EVENT_TABLE()
};

BEGIN_EVENT_TABLE(Settings, wxPanel)
  EVT_UPDATE_UI(id_args, Settings::on_args_update)
  EVT_UPDATE_UI(id_file_name, Settings::on_file_name_update)
  EVT_BUTTON(id_orchestra, Settings::on_orchestra)
  EVT_BUTTON(id_score, Settings::on_score)
  EVT_BUTTON(id_output, Settings::on_output)
  EVT_BUTTON(id_clear_orchestra, Settings::on_clear_orchestra)
  EVT_BUTTON(id_clear_score, Settings::on_clear_score)
  EVT_BUTTON(id_clear_output, Settings::on_clear_output)
  EVT_CHECKBOX(id_no_sound, Settings::on_no_sound)
  EVT_CHECKBOX(id_rewrite_header, Settings::on_rewrite_header)
  EVT_CHECKBOX(id_peak_chunks, Settings::on_peak_chunks)
  EVT_RADIOBOX(id_format, Settings::on_format)
  EVT_RADIOBOX(id_size, Settings::on_size)
  EVT_RADIOBOX(id_graphics, Settings::on_graphics)
END_EVENT_TABLE()

void Settings::stack_two(wxPanel *panel, wxWindow *check, wxWindow *radio)
{
  wxBoxSizer *sizer = new wxBoxSizer(wxVERTICAL);
  sizer->Add(check, 0, wxALL | wxALIGN_CENTRE, 2);
  sizer->Add(radio, 0, wxALL | wxALIGN_CENTRE, 2);
  sizer->Add(new wxWindow(panel, -1), 1, wxEXPAND | wxALL, 2);
  panel->SetSizer(sizer);
  sizer->SetSizeHints(panel);
}

Settings::Settings(Args *args, wxFileName *file_name, wxWindow *parent)
  : wxPanel(parent), m_args(args), m_file_name(file_name)
{
  wxBoxSizer *sizer = new wxBoxSizer(wxVERTICAL);
  wxStaticText *text;

  // title
  text = new wxStaticText(this, -1, wxT("Configure a Csound Task"));
  sizer->Add(text, 0, wxALL | wxALIGN_CENTRE, 2);
  text = new wxStaticText(this, -1, wxT(""));
  sizer->Add(text, 0, wxALL, 2);

  // files
  wxPanel *files = new wxPanel(this);
  wxFlexGridSizer *flex = new wxFlexGridSizer(6, 2, 2);
  flex->AddGrowableCol(2);
  wxButton *button;

  flex->Add(new wxWindow(files, -1), 0, wxEXPAND | wxALL, 2);
  text = new wxStaticText(files, -1, wxT("Orchestra: "));
  flex->Add(text, 0, wxALL | wxEXPAND, 2);
  m_orchestra = new wxStaticText(files, -1, empty_file_name);
  flex->Add(m_orchestra, 1, wxALL | wxEXPAND, 2);
  button = new wxButton(files, id_orchestra, wxT("Change"));
  flex->Add(button, 0, wxALL, 2);
  button = new wxButton(files, id_clear_orchestra, wxT("Clear"));
  flex->Add(button, 0, wxALL, 2);
  flex->Add(new wxWindow(files, -1), 0, wxEXPAND | wxALL, 2);

  flex->Add(new wxWindow(files, -1), 0, wxEXPAND | wxALL, 2);
  text = new wxStaticText(files, -1, wxT("Score: "));
  flex->Add(text, 0, wxALL | wxEXPAND, 2);
  m_score = new wxStaticText(files, -1, empty_file_name);
  flex->Add(m_score, 1, wxALL | wxEXPAND, 2);
  button = new wxButton(files, id_score, wxT("Change"));
  flex->Add(button, 0, wxALL, 2);
  button = new wxButton(files, id_clear_score, wxT("Clear"));
  flex->Add(button, 0, wxALL, 2);
  flex->Add(new wxWindow(files, -1), 0, wxEXPAND | wxALL, 2);

  flex->Add(new wxWindow(files, -1), 0, wxEXPAND | wxALL, 2);
  text = new wxStaticText(files, -1, wxT("Output: "));
  flex->Add(text, 0, wxALL | wxEXPAND, 2);
  m_output = new wxStaticText(files, -1, empty_file_name);
  flex->Add(m_output, 1, wxALL | wxEXPAND, 2);
  button = new wxButton(files, id_output, wxT("Change"));
  flex->Add(button, 0, wxALL, 2);
  button = new wxButton(files, id_clear_output, wxT("Clear"));
  flex->Add(button, 0, wxALL, 2);
  flex->Add(new wxWindow(files, -1), 0, wxEXPAND | wxALL, 2);

  files->SetSizer(flex);
  flex->SetSizeHints(files);
  sizer->Add(files, 1, wxALL | wxEXPAND, 2);

  // switches
  const wxString format_choices[] = {
    default_format,
    aiff_format,
    wave_format,
    ircam_format,
    raw_format,
  };
  const wxString size_choices[] = {
    default_size,
    eight_unsigned_size,
    eight_signed_size,
    sixteen_size,
    twenty_four_size,
    thirty_two_size,
    float_size,
  };
  const wxString graphics_choices[] = {
    full_graphics,
    postscript_graphics,
    ascii_graphics,
    no_graphics,
  };

  wxPanel *switches = new wxPanel(this);
  wxBoxSizer *switches_sizer = new wxBoxSizer(wxHORIZONTAL);

  switches_sizer->Add(new wxWindow(switches, -1), 1, wxEXPAND | wxALL, 2);
  wxPanel *panel;

  panel = new wxPanel(switches);
  m_no_sound = new wxCheckBox(panel, id_no_sound, wxT("No Sound"));
  m_format = new wxRadioBox(panel, id_format, wxT("Format"),
			    wxDefaultPosition, wxDefaultSize,
			    WXSIZEOF(format_choices), format_choices, 1);
  stack_two(panel, m_no_sound, m_format);
  switches_sizer->Add(panel, 0, wxALL, 2);
  switches_sizer->Add(new wxWindow(switches, -1), 1, wxEXPAND | wxALL, 2);

  panel = new wxPanel(switches);
  m_size = new wxRadioBox(panel, id_size, wxT("Size"),
			  wxDefaultPosition, wxDefaultSize,
			  WXSIZEOF(size_choices), size_choices, 1);
  m_rewrite_header = new wxCheckBox(panel, id_rewrite_header,
				    wxT("Rewrite Header"));
  stack_two(panel, m_rewrite_header, m_size);
  switches_sizer->Add(panel, 0, wxALL, 2);
  switches_sizer->Add(new wxWindow(switches, -1), 1, wxEXPAND | wxALL, 2);

  panel = new wxPanel(switches);
  m_graphics = new wxRadioBox(panel, id_graphics, wxT("Graphics"),
			      wxDefaultPosition, wxDefaultSize,
			      WXSIZEOF(graphics_choices), graphics_choices, 1);
  m_peak_chunks = new wxCheckBox(panel, id_peak_chunks, wxT("Peak Chunks"));
  stack_two(panel, m_peak_chunks, m_graphics);
  switches_sizer->Add(panel, 0, wxALL, 2);
  switches_sizer->Add(new wxWindow(switches, -1), 1, wxEXPAND | wxALL, 2);

  switches->SetSizer(switches_sizer);
  switches_sizer->SetSizeHints(switches);
  sizer->Add(switches, 0, wxEXPAND | wxALL, 2);

  sizer->Add(new wxWindow(this, -1), 1, wxEXPAND | wxALL, 2);
  m_file = new wxStaticText(this, -1, wxT(""));
  sizer->Add(m_file, 0, wxEXPAND | wxALL, 2);
  m_command = new wxStaticText(this, -1, wxT(""));
  sizer->Add(m_command, 0, wxEXPAND | wxALL, 2);

  SetSizer(sizer);
  sizer->SetSizeHints(this);
  do_args_update();
}

void Settings::on_args_update(wxUpdateUIEvent& WXUNUSED(event))
{
  do_args_update();
}

void Settings::on_file_name_update(wxUpdateUIEvent& WXUNUSED(event))
{
  wxString command;
  if (m_file_name->IsOk()) {
    command.Append(wxT("File: "));
    command.Append(m_file_name->GetFullPath());
  }
  m_file->SetLabel(command);
  do_args_update();
}

void Settings::do_args_update()
{
  wxFileName file_name(*m_file_name);
  if (!file_name.IsOk())	// Use cwd as base file for making relative
    file_name = wxFileName(wxFileName::GetCwd(), wxT(""));

  if (m_args->get_orchestra().IsOk())
    m_orchestra->SetLabel(m_args->get_orchestra(file_name).GetFullPath());
  else
    m_orchestra->SetLabel(empty_file_name);
  if (m_args->get_score().IsOk())
    m_score->SetLabel(m_args->get_score(file_name).GetFullPath());
  else
    m_score->SetLabel(empty_file_name);
  if (m_args->get_output().IsOk())
    m_output->SetLabel(m_args->get_output(file_name).GetFullPath());
  else
    m_output->SetLabel(empty_file_name);

  switch (m_args->get_format()) {
  case AIFF_FORMAT:
    m_format->SetStringSelection(aiff_format);
    break;
  case WAVE_FORMAT:
    m_format->SetStringSelection(wave_format);
    break;
  case IRCAM_FORMAT:
    m_format->SetStringSelection(ircam_format);
    break;
  case RAW_FORMAT:
    m_format->SetStringSelection(raw_format);
    break;
  default:
    m_format->SetStringSelection(default_format);
  }

  switch (m_args->get_output_size()) {
  case UNSIGNED_EIGHT_SIZE:
    m_size->SetStringSelection(eight_unsigned_size);
    break;
  case SIGNED_EIGHT_SIZE:
    m_size->SetStringSelection(eight_signed_size);
    break;
  case SHORT_SIZE:
    m_size->SetStringSelection(sixteen_size);
    break;
  case TWENTY_FOUR_SIZE:
    m_size->SetStringSelection(twenty_four_size);
    break;
  case THIRTY_TWO_SIZE:
    m_size->SetStringSelection(thirty_two_size);
    break;
  case FLOAT_SIZE:
    m_size->SetStringSelection(float_size);
    break;
  default:
    m_size->SetStringSelection(default_size);
  }

  switch (m_args->get_graphics_option()) {
  case ASCII_GRAPHICS:
    m_graphics->SetStringSelection(ascii_graphics);
    break;
  case POSTSCRIPT_GRAPHICS:
    m_graphics->SetStringSelection(postscript_graphics);
    break;
  case NO_GRAPHICS:
    m_graphics->SetStringSelection(no_graphics);
    break;
  default:
    m_graphics->SetStringSelection(full_graphics);
  }

  m_no_sound->SetValue(m_args->get_no_sound());
  m_rewrite_header->SetValue(m_args->get_rewrite_header());
  m_peak_chunks->SetValue(!m_args->get_no_peak_chunks());

  wxString command(wxT("Command: "));
  command.Append(enquote(m_args->get_command_line(file_name)));
  m_command->SetLabel(command);
}

void Settings::on_orchestra(wxCommandEvent& WXUNUSED(event))
{
  wxFileDialog dialog(this, wxT("Orchestra file"),
		      m_file_name->GetPath(),
		      m_args->get_orchestra().GetFullName(),
		      orchestra_wildcard,
		      wxOPEN | wxHIDE_READONLY);
  if (dialog.ShowModal() == wxID_OK)
    m_args->set_orchestra(dialog.GetPath());
}

void Settings::on_score(wxCommandEvent& WXUNUSED(event))
{
  wxString name;
  wxFileName name_template = m_args->get_score();
  if (name_template.IsOk())
    name = name_template.GetFullName();
  else {
    name_template = m_args->get_orchestra();
    if (name_template.IsOk()) {	// Use orchestra to construct name
      name_template.SetExt(score_ext);
      name = name_template.GetFullName();
    }
  }
  wxFileDialog dialog(this, wxT("Score file"),
		      m_file_name->GetPath(),
		      name, score_wildcard,
		      wxOPEN | wxHIDE_READONLY);
  if (dialog.ShowModal() == wxID_OK)
    m_args->set_score(dialog.GetPath());
}

void Settings::on_output(wxCommandEvent& WXUNUSED(event))
{
  Format format = m_args->get_format();
  wxString name;
  wxFileName name_template = m_args->get_output();
  if (!name_template.IsOk())	// Use orchestra as template
    name_template = m_args->get_orchestra();
  if (name_template.IsOk()) { // Use format to set extension
    name_template.SetExt(format_ext[format]);
    name = name_template.GetFullName();
  }
  wxFileDialog dialog(this, wxT("Output file"),
		      m_file_name->GetPath(),
		      name, format_wildcard[format],
		      wxSAVE | wxOVERWRITE_PROMPT);
  if (dialog.ShowModal() == wxID_OK)
    m_args->set_output(dialog.GetPath());
}

void Settings::on_clear_orchestra(wxCommandEvent& WXUNUSED(event))
{
  wxFileName fresh;
  m_args->set_orchestra(fresh);
}

void Settings::on_clear_score(wxCommandEvent& WXUNUSED(event))
{
  wxFileName fresh;
  m_args->set_score(fresh);
}

void Settings::on_clear_output(wxCommandEvent& WXUNUSED(event))
{
  wxFileName fresh;
  m_args->set_output(fresh);
}

void Settings::on_no_sound(wxCommandEvent& WXUNUSED(event))
{
  m_args->set_no_sound(m_no_sound->GetValue());
}

void Settings::on_rewrite_header(wxCommandEvent& WXUNUSED(event))
{
  m_args->set_rewrite_header(m_rewrite_header->GetValue());
}

void Settings::on_peak_chunks(wxCommandEvent& WXUNUSED(event))
{
  m_args->set_no_peak_chunks(!m_peak_chunks->GetValue());
}

void Settings::on_format(wxCommandEvent& WXUNUSED(event))
{
  wxString item = m_format->GetStringSelection();
  if (item.IsSameAs(aiff_format))
    m_args->set_format(AIFF_FORMAT);
  else if (item.IsSameAs(wave_format))
    m_args->set_format(WAVE_FORMAT);
  else if (item.IsSameAs(ircam_format))
    m_args->set_format(IRCAM_FORMAT);
  else if (item.IsSameAs(raw_format))
    m_args->set_format(RAW_FORMAT);
  else if (item.IsSameAs(default_format))
    m_args->set_format(NO_FORMAT);
}

void Settings::on_size(wxCommandEvent& WXUNUSED(event))
{
  wxString item = m_size->GetStringSelection();
  if (item.IsSameAs(eight_unsigned_size))
    m_args->set_output_size(UNSIGNED_EIGHT_SIZE);
  else if (item.IsSameAs(eight_signed_size))
    m_args->set_output_size(SIGNED_EIGHT_SIZE);
  else if (item.IsSameAs(sixteen_size))
    m_args->set_output_size(SHORT_SIZE);
  else if (item.IsSameAs(twenty_four_size))
    m_args->set_output_size(TWENTY_FOUR_SIZE);
  else if (item.IsSameAs(thirty_two_size))
    m_args->set_output_size(THIRTY_TWO_SIZE);
  else if (item.IsSameAs(float_size))
    m_args->set_output_size(FLOAT_SIZE);
  else if (item.IsSameAs(default_size))
    m_args->set_output_size(NO_SIZE);
}

void Settings::on_graphics(wxCommandEvent& WXUNUSED(event))
{
  wxString item = m_graphics->GetStringSelection();
  if (item.IsSameAs(full_graphics))
    m_args->set_graphics_option(FULL_GRAPHICS);
  else if (item.IsSameAs(postscript_graphics))
    m_args->set_graphics_option(POSTSCRIPT_GRAPHICS);
  else if (item.IsSameAs(ascii_graphics))
    m_args->set_graphics_option(ASCII_GRAPHICS);
  else if (item.IsSameAs(no_graphics))
    m_args->set_graphics_option(NO_GRAPHICS);
}

class Main_frame;

// Declaration of the model for the main frame

class Main_model
{
public:
  void init(Main_frame *, wxNotebook *, Settings *, Transcript *, Canvas *);
  void do_new();
  void do_open();
  void do_open_unified();
  void really_do_open(const wxFileName&);
  void do_save();
  void do_save_as();
  void do_quit();
  void do_run();
  void do_stop();
  void do_clear();
  void do_about();
  void do_list_opcodes();
  void do_usage();
  void update_status();
  void show_settings();
  void show_transcript();
  void show_canvas();
  void show_messages();
  void note_canves();
  void note_halt();
  void note_transcript_text(const wxString&);
  void note_clear_curves();
  void note_draw_curves();
  Args *get_args();
  wxFileName *get_file_name();
private:
  bool m_idle;			// Is the worker idle?
  Args m_args;
  wxFileName m_file_name;	// Name of configuration file
  Main_frame *m_frame;
  wxNotebook *m_notebook;
  Settings *m_settings;
  Transcript *m_transcript;
  Canvas *m_canvas;
  void set_file_name(const wxFileName&);
  void really_do_save(const wxFileName&);
};

// The class that enables drag and drop of CSC files into wxCsound.
class Config_drop: public wxFileDropTarget
{
public:
  Config_drop(Main_model *model) 
    : m_model(model)
  {
  }
  bool OnDropFiles(wxCoord, wxCoord, const wxArrayString&);
private:
  Main_model *m_model;
};

bool Config_drop::OnDropFiles(wxCoord x, wxCoord y, 
			      const wxArrayString& filenames)
{
  if (filenames.GetCount() != 1)
    return FALSE;
  wxString& filename = filenames.Item(0);
  m_model->really_do_open(filename);
  return TRUE;
}

// The main frame

class Main_frame : public wxFrame
{
public:
  Main_frame(Table *, const wxString&, const wxPoint&);

  void on_new(wxCommandEvent&);
  void on_open(wxCommandEvent&);
  void on_open_unified(wxCommandEvent&);
  void on_save(wxCommandEvent&);
  void on_save_as(wxCommandEvent&);
  void on_quit(wxCommandEvent&);
  void on_run(wxCommandEvent&);
  void on_stop(wxCommandEvent&);
  void on_clear(wxCommandEvent&);
  void on_about(wxCommandEvent&);
  void on_viewer(wxCommandEvent&);
  void on_list_opcodes(wxCommandEvent&);
  void on_usage(wxCommandEvent&);
  void on_settings(wxCommandEvent&);
  void on_transcript(wxCommandEvent&);
  void on_canvas(wxCommandEvent&);
  void on_messages(wxCommandEvent&);
  void on_halt(wxUpdateUIEvent&);
  void on_transcript_text(wxUpdateUIEvent&);
  void on_clear_curves(wxUpdateUIEvent&);
  void on_draw_curves(wxUpdateUIEvent&);
  void open_file(const wxFileName&);

private:
  static void add_one_to(wxWindow *, wxWindow *);
  Main_model m_model;
  Viewer m_viewer;
  DECLARE_EVENT_TABLE()
};

BEGIN_EVENT_TABLE(Main_frame, wxFrame)
  EVT_MENU(id_new, Main_frame::on_new)
  EVT_MENU(id_open, Main_frame::on_open)
  EVT_MENU(id_open_unified, Main_frame::on_open_unified)
  EVT_MENU(id_save, Main_frame::on_save)
  EVT_MENU(id_save_as, Main_frame::on_save_as)
  EVT_MENU(id_quit, Main_frame::on_quit)
  EVT_MENU(id_run, Main_frame::on_run)
  EVT_MENU(id_stop, Main_frame::on_stop)
  EVT_MENU(id_clear, Main_frame::on_clear)
  EVT_MENU(id_about, Main_frame::on_about)
  EVT_MENU(id_viewer, Main_frame::on_viewer)
  EVT_MENU(id_list_opcodes, Main_frame::on_list_opcodes)
  EVT_MENU(id_usage, Main_frame::on_usage)
  EVT_MENU(id_settings, Main_frame::on_settings)
  EVT_MENU(id_transcript, Main_frame::on_transcript)
  EVT_MENU(id_canvas, Main_frame::on_canvas)
  EVT_MENU(id_messages, Main_frame::on_messages)
  EVT_UPDATE_UI(id_halt, Main_frame::on_halt)
  EVT_UPDATE_UI(id_text, Main_frame::on_transcript_text)
  EVT_UPDATE_UI(id_clear_curves, Main_frame::on_clear_curves)
  EVT_UPDATE_UI(id_draw_curves, Main_frame::on_draw_curves)
END_EVENT_TABLE()

void Main_frame::add_one_to(wxWindow *parent, wxWindow *child)
{
  wxBoxSizer *sizer = new wxBoxSizer(wxVERTICAL);
  sizer->Add(child,
	     1,			// make vertically stretchable
	     wxEXPAND | wxALL,	// make horizontally stretchable
				//   and make border all around
	     2);		// set border width
  parent->SetSizer(sizer);
  sizer->SetSizeHints(parent);
}

Main_frame::Main_frame(Table *table, const wxString& title, const wxPoint& pos)
  : wxFrame((wxFrame *)NULL, -1, title, pos, wxDefaultSize),
    m_viewer(this)
{
  wxMenuBar *menuBar = new wxMenuBar;
  wxMenu *menu;

  menu = new wxMenu;
  menu->Append(id_new, wxT("&New"),
	       wxT("Clear settings"));
  menu->Append(id_open, wxT("&Open"),
	       wxT("Load a Csound Configuration"));
  menu->Append(id_open_unified, wxT("Open &Unified"),
	       wxT("Create a configuration for a Csound Unified file"));
  menu->Append(id_save, wxT("&Save"),
	       wxT("Save a Csound Configuration in the current file"));
  menu->Append(id_save_as, wxT("Save &As"),
	       wxT("Save a Csound Configuration in a specified file"));
  menu->Append(id_quit, wxT("&Quit"),
	       wxT("Quit application"));
  menuBar->Append(menu, wxT("&File"));

  // Maybe there should be an edit menu bar entry with menu items for
  // copy and paste.

  menu = new wxMenu;
  menu->Append(id_run, wxT("&Run Task"),
	       wxT("Run a Csound task"));
  menu->Append(id_stop, wxT("&Stop Task"),
	       wxT("Stop a running Csound task"));
  menu->Append(id_clear, wxT("&Clear Transcript"),
	       wxT("Clear the transcript window"));
  menuBar->Append(menu, wxT("&Action"));

  menu = new wxMenu;
  menu->Append(id_settings, wxT("&Settings"),
	       wxT("Switch to the settings window"));
  menu->Append(id_transcript, wxT("&Transcript"),
	       wxT("Switch to the transcript window"));
  menu->Append(id_canvas, wxT("&Plots"),
	       wxT("Switch to the plots window"));
  menu->Append(id_messages, wxT("&Messages"),
	       wxT("Switch to the messages window"));
  menuBar->Append(menu, wxT("&Window"));

  menu = new wxMenu;
  menu->Append(id_about, wxT("&About"),
	       wxT("Show ") wxT(PACKAGE_NAME) wxT(" version"));
  menu->Append(id_viewer, wxT("&Overview"),
	       wxT("Show ") wxT(PACKAGE_NAME) wxT(" overview"));
  menu->Append(id_list_opcodes, wxT("&List Opcodes"),
	       wxT("List opcodes in Csound"));
  menu->Append(id_usage, wxT("&CSC Syntax"),
	       wxT("Show Csound Configuration file syntax"));
  menuBar->Append(menu, wxT("&Help"));

  SetMenuBar(menuBar);

  // Status Bar

  CreateStatusBar(2);
  int bar_widths[] = { -5, -1 };
  SetStatusWidths(2, bar_widths);

  // Icons

  wxIconBundle bundle;
  bundle.AddIcon(wxcsound128_xpm);
  bundle.AddIcon(wxcsound112_xpm);
  bundle.AddIcon(wxcsound96_xpm);
  bundle.AddIcon(wxcsound80_xpm);
  bundle.AddIcon(wxcsound64_xpm);
  bundle.AddIcon(wxcsound48_xpm);
  bundle.AddIcon(wxcsound32_xpm);
  bundle.AddIcon(wxcsound16_xpm);
  SetIcons(bundle);

  // Keep this code in sync with enum pages

  wxNotebook *notebook = new wxNotebook(this, -1);
  wxNotebookSizer *nbs = new wxNotebookSizer(notebook);

  Args *args = m_model.get_args();
  wxFileName *file_name = m_model.get_file_name();
  Settings *settings = new Settings(args, file_name, notebook);
  notebook->AddPage(settings, wxString(wxT("Settings")), TRUE);
  args->set_handler(settings);	// Direct update events to settings
  args->set_id(id_args);	// and use id_args as the identifier

  Transcript *transcript = new Transcript(notebook);
  notebook->AddPage(transcript, wxString(wxT("Transcript")), FALSE);

  Canvas *canvas = new Canvas(table, notebook);
  notebook->AddPage(canvas, wxString(wxT("Plots")), FALSE);

  wxPanel *messages = new wxPanel(notebook);
  notebook->AddPage(messages, wxString(wxT("Messages")), FALSE);
  wxTextCtrl *text = new wxTextCtrl(messages, -1, wxT(""),
			  wxDefaultPosition, wxDefaultSize,
			  wxTE_MULTILINE | wxTE_READONLY);
  add_one_to(messages, text);
  wxLog::SetActiveTarget(new wxLogTextCtrl(text));
  wxMessageOutput::Set(new wxMessageOutputLog);

  nbs->SetSizeHints(notebook);
  add_one_to(this, notebook);

  m_model.init(this, notebook, settings, transcript, canvas);

  SetDropTarget(new Config_drop(&m_model));
}

// Frame event handlers

void Main_frame::on_new(wxCommandEvent& WXUNUSED(event))
{
  m_model.do_new();
}

void Main_frame::on_open(wxCommandEvent& WXUNUSED(event))
{
  m_model.do_open();
}

void Main_frame::on_open_unified(wxCommandEvent& WXUNUSED(event))
{
  m_model.do_open_unified();
}

void Main_frame::on_save(wxCommandEvent& WXUNUSED(event))
{
  m_model.do_save();
}

void Main_frame::on_save_as(wxCommandEvent& WXUNUSED(event))
{
  m_model.do_save_as();
}

void Main_frame::on_quit(wxCommandEvent& WXUNUSED(event))
{
  m_model.do_quit();
}

void Main_frame::on_run(wxCommandEvent& WXUNUSED(event))
{
  m_model.do_run();
}

void Main_frame::on_settings(wxCommandEvent& WXUNUSED(event))
{
  m_model.show_settings();
}

void Main_frame::on_transcript(wxCommandEvent& WXUNUSED(event))
{
  m_model.show_transcript();
}

void Main_frame::on_canvas(wxCommandEvent& WXUNUSED(event))
{
  m_model.show_canvas();
}

void Main_frame::on_messages(wxCommandEvent& WXUNUSED(event))
{
  m_model.show_messages();
}

void Main_frame::on_stop(wxCommandEvent& WXUNUSED(event))
{
  m_model.do_stop();
}

void Main_frame::on_clear(wxCommandEvent& WXUNUSED(event))
{
  m_model.do_clear();
}

void Main_frame::on_about(wxCommandEvent& WXUNUSED(event))
{
  m_model.do_about();
}

void Main_frame::on_viewer(wxCommandEvent& WXUNUSED(event))
{
  m_viewer.show();
}

void Main_frame::on_list_opcodes(wxCommandEvent& WXUNUSED(event))
{
  m_model.do_list_opcodes();
}

void Main_frame::on_usage(wxCommandEvent& WXUNUSED(event))
{
  m_model.do_usage();
}

void Main_frame::on_halt(wxUpdateUIEvent& WXUNUSED(event))
{
  m_model.note_halt();
}

void Main_frame::on_transcript_text(wxUpdateUIEvent& event)
{
  m_model.note_transcript_text(event.GetText());
}

void Main_frame::on_draw_curves(wxUpdateUIEvent& WXUNUSED(event))
{
  m_model.note_draw_curves();
}

void Main_frame::on_clear_curves(wxUpdateUIEvent& WXUNUSED(event))
{
  m_model.note_clear_curves();
}

void Main_frame::open_file(const wxFileName& file_name)
{
  m_model.really_do_open(file_name);
}

// The implementation of the main model

void Main_model::init(Main_frame *frame, wxNotebook *notebook,
		      Settings *settings, Transcript *transcript,
		      Canvas *canvas) {
  m_idle = TRUE;
  m_frame = frame;
  m_notebook = notebook;
  m_settings = settings;
  m_transcript = transcript;
  m_canvas = canvas;
  update_status();
}

Args *Main_model::get_args()
{
  return &m_args;
}

wxFileName *Main_model::get_file_name()
{
  return &m_file_name;
}

void Main_model::set_file_name(const wxFileName& file_name)
{
  if (m_file_name.SameAs(file_name))
    return;
  m_file_name = file_name;
  wxUpdateUIEvent ev(id_file_name);
  m_settings->AddPendingEvent(ev);
}

void Main_model::do_new()
{
  m_args.clear();
  wxFileName file_name;
  set_file_name(file_name);
}

void Main_model::do_open()
{
  wxFileDialog dialog(m_frame, wxT("Open configuration file"),
		      m_file_name.GetPath(), m_file_name.GetFullName(),
		      config_wildcard,
		      wxOPEN | wxHIDE_READONLY | wxCHANGE_DIR);
  if (dialog.ShowModal() == wxID_OK)
    really_do_open(dialog.GetPath());
}

void Main_model::do_open_unified()
{
  wxFileDialog dialog(m_frame, wxT("Open Csound Unified file"),
		      m_file_name.GetPath(), m_file_name.GetFullName(),
		      csd_wildcard, wxOPEN | wxHIDE_READONLY);
  if (dialog.ShowModal() == wxID_OK) {
    do_new();
    m_args.set_orchestra(dialog.GetPath());
  }
}

const int buffer_size = 1 << 13;

void Main_model::really_do_open(const wxFileName& file_name)
{
  wxFFile file;
  if (!file.Open(file_name.GetFullPath().c_str(), "r")) {
    wxString message(wxT("File open for read error for: "));
    message.Append(file_name.GetFullPath());
    wxMessageBox(message, m_frame->GetTitle(),
		 wxOK | wxICON_ERROR);
    return;
  }
  char buffer[buffer_size + 1];
  size_t n = file.Read(buffer, buffer_size);
  if (n < 0) {
    wxString message(wxT("File read error for: "));
    message.Append(file_name.GetFullPath());
    wxMessageBox(message, m_frame->GetTitle(),
		 wxOK | wxICON_ERROR);
    return;
  }
  buffer[n] = 0;
  wxString line(buffer, wxConvUTF8);
  wxArrayString args = dequote(line);
  if (!m_args.parse(file_name, args)) {
    set_file_name(file_name);
    wxLogMessage(wxT("Opened: %s"), file_name.GetFullPath().c_str());
  }
  else {
    wxString message(wxT("Parse error in contents of: "));
    message.Append(file_name.GetFullPath());
    wxLogMessage(wxT("%s"), message.c_str());
    wxMessageBox(message, m_frame->GetTitle(),
		 wxOK | wxICON_ERROR);
  }
}

void Main_model::do_save()
{
  if (!m_file_name.IsOk())
    do_save_as();
  else
    really_do_save(m_file_name);
}

void Main_model::do_save_as()
{
  wxFileDialog dialog(m_frame, wxT("Save configuration file"),
		      m_file_name.GetPath(), m_file_name.GetFullName(),
		      config_wildcard,
		      wxSAVE | wxOVERWRITE_PROMPT | wxCHANGE_DIR);
  if (dialog.ShowModal() == wxID_OK)
    really_do_save(dialog.GetPath());
}

void Main_model::really_do_save(const wxFileName& file_name)
{
  if (!file_name.IsOk())
    return;
  wxFFile file;
  if (!file.Open(file_name.GetFullPath().c_str(), "w")) {
    wxString message(wxT("File open for write error for: "));
    message.Append(file_name.GetFullPath());
    wxMessageBox(message, m_frame->GetTitle(),
		 wxOK | wxICON_ERROR);
    return;
  }
  if (!file.Write(enquote(m_args.get_command_line(file_name)), wxConvUTF8)) {
    wxString message(wxT("File write error for: "));
    message.Append(file_name.GetFullPath());
    wxMessageBox(message, m_frame->GetTitle(),
		 wxOK | wxICON_ERROR);
    return;
  }
  file.Write("\n", 1);
  set_file_name(file_name);
  wxLogMessage(wxT("Saved: %s"), file_name.GetFullPath().c_str());
}

// Kill worker, wait until it stops, then kill the main frame.
void Main_model::do_quit()
{
  Worker::get_worker()->kill_worker();
  m_frame->Close(TRUE);
}

// Start worker and switch to transcript window.
void Main_model::do_run()
{
  wxFileName cwd(wxFileName::GetCwd(), wxT(""));
  wxArrayString cmdline = m_args.get_command_line(cwd);
  show_transcript();
  m_idle = Worker::get_worker()->start_worker(cmdline) < 0;
  update_status();
  if (!m_idle)
    wxLogMessage(wxT("Started task: %s"), enquote(cmdline).c_str());
}

// Request that the worker stop and switch to transcript window.
void Main_model::do_stop()
{
  if (!m_idle)
    wxLogMessage(wxT("Stopping current task"));
  show_transcript();
  Worker::get_worker()->set_abort_pending();
}

// Clear the transcript window.
void Main_model::do_clear()
{
  m_transcript->clear();
  show_transcript();
}

// Show help message.
void Main_model::do_about()
{
  wxString message(m_frame->GetTitle());
  message.Append(wxT(" with Csound "));
  message.Append(Synthesizer::get_version());
  wxMessageBox(message, wxT("About ") wxT(PACKAGE_NAME),
	       wxOK | wxICON_INFORMATION);
}

void Main_model::do_list_opcodes()
{
  wxArrayString cmdline;
  cmdline.Add(wxT("csound"));
  cmdline.Add(wxT("-z"));
  show_transcript();
  m_idle = Worker::get_worker()->start_worker(cmdline) < 0;
  update_status();
}

void Main_model::do_usage()
{
  m_args.usage();
  show_messages();
}

// Update the status line based on the idle flag.
void Main_model::update_status()
{
  if (m_idle)
    m_frame->SetStatusText(wxT("Ready"), 1);
  else
    m_frame->SetStatusText(wxT("Running"), 1);
}

void Main_model::show_settings()
{
  m_notebook->SetSelection(page_settings);
}

void Main_model::show_transcript()
{
  m_notebook->SetSelection(page_transcript);
}

void Main_model::show_canvas()
{
  m_notebook->SetSelection(page_canvas);
}

void Main_model::show_messages()
{
  m_notebook->SetSelection(page_messages);
}

// The worker halted, so change idle flag.
void Main_model::note_halt()
{
  m_idle = TRUE;
  update_status();
  wxLogMessage(wxT("Task halted"));
}

// The worker sent text for the transcript window.
void Main_model::note_transcript_text(const wxString& text)
{
  m_transcript->append_text(text);
}

// The worker cleared the curves on the canvas.
void Main_model::note_clear_curves()
{
  m_canvas->do_paint();
}

// The worker changed the curves to be drawn on the canvas.
void Main_model::note_draw_curves()
{
  show_canvas();
  m_canvas->do_paint();
}

// The implementation of the application

IMPLEMENT_APP(Main)

bool Main::OnInit()
{
  wxString title(wxT(PACKAGE_STRING));
  wxChar *file_name = 0;
  switch (argc) {
  case 1:
    break;
  case 2:
    file_name = argv[1];
    if (*file_name != '-')
      break;
    // fall thru okay
  default:
    wxString msg;
    msg.Printf(wxT("%s\nUsage: %s [csound-config-file]"),
	       title.c_str(), argv[0]);
    puts(msg.mb_str(wxConvUTF8));
    return FALSE;
  }
  Main_frame *frame = new Main_frame(&m_table, title,
				     wxPoint(app_x_pos, app_y_pos));
  frame->Show(TRUE);
  SetTopWindow(frame);
  if (file_name && *file_name) {
    wxFileName name(file_name);
    name.MakeAbsolute();
    frame->open_file(name);
  }
  return TRUE;
}

// Thread safe worker thread halt notifier.
void Main::post_halt(int rc)
{
  wxUpdateUIEvent ev(id_halt);
  GetTopWindow()->AddPendingEvent(ev);
}

// Thread safe text for transcript notifier.
void Main::post_transcript_text(const wxString& text)
{
  wxUpdateUIEvent ev(id_text);
  ev.SetText(text);
  GetTopWindow()->AddPendingEvent(ev);
}

// Thread safe curves cleared notifier.
void Main::post_clear_plots()
{
  m_table.clear();
  wxUpdateUIEvent ev(id_clear_curves);
  GetTopWindow()->AddPendingEvent(ev);
}

// Thread safe curve added notifier.
void Main::post_plot_curve(Curve *curve)
{
  m_table.add(curve);
  wxUpdateUIEvent ev(id_draw_curves);
  GetTopWindow()->AddPendingEvent(ev);
}
