/*
 * Csound worker and thread--the thread that does the rendering
 *
 * Copyright (c) 2003 by John D. Ramsdell. All rights reserved.
 *
 * This software is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <wx/wxprec.h>

#ifndef WX_PRECOMP
#include <wx/app.h>
#include <wx/thread.h>
#endif

#include "synthesizer.hpp"
#include "curve.hpp"
#include "table.hpp"
#include "worker.hpp"
#include "main.hpp"

// The worker synchronizes access to the abort pending flag and houses
// the worker thread.  It also routes poll events generated by the
// worker thread to the worker thread object.

class Worker_thread : public wxThread
{
public:
  Worker_thread(Synthesizer *m_synthesizer, const wxArrayString&);
  ExitCode Entry();
  bool poll_events();
private:
  Synthesizer *m_synthesizer;
  wxArrayString m_args;
};

Worker *Worker::s_worker = 0;

// Ensure there is only one worker
Worker *Worker::get_worker()
{
  if (!s_worker)
    s_worker = new Worker();
  return s_worker;
}

extern "C" int wxcsound_poll_events(void *csound)
{
  return Worker::get_worker()->poll_events();
}

Worker::Worker()
  : m_thread(0), m_abort_pending(FALSE)
{
  m_synthesizer.set_yield_callback(wxcsound_poll_events);
}

void Worker::set_abort_pending()
{
  m_abort_critical_section.Enter();
  m_abort_pending = TRUE;
  m_abort_critical_section.Leave();
}

bool Worker::is_abort_pending()
{
  m_abort_critical_section.Enter();
  bool result = m_abort_pending;
  if (result)
    m_abort_pending = FALSE;
  m_abort_critical_section.Leave();
  return result;
}

// Returns -1 when the worker is already rendering
int Worker::start_worker(const wxArrayString& args)
{
  Worker_thread *thread = m_thread;
  if (thread) {
    if (thread->IsAlive())	// Thread active
      return -1;
    else {
      thread->Wait();		// Last thread done
      delete thread;		// Recover resources
    }
  }
  is_abort_pending();		// Clear flag at start of run
  // Make new thread
  thread = new Worker_thread(&m_synthesizer, args);
  m_thread = thread;
  thread->Create();
  thread->Run();
  return 0;
}

// Stops a rendering and returns when the thread stops
void Worker::kill_worker()
{
  set_abort_pending();
  Worker_thread *thread = m_thread;
  if (thread) {
    thread->Wait();
    delete thread;
    m_thread = 0;
  }
}

// Called in the worker thread
int Worker::poll_events()
{
  if (is_abort_pending())
    return 0;
  Worker_thread *thread = m_thread;
  if (!thread)		// This should never happen
    return 1;
  else
    return thread->poll_events();
}

Worker_thread::Worker_thread(Synthesizer *synthesizer,
			     const wxArrayString& args)
  : wxThread(wxTHREAD_JOINABLE), m_synthesizer(synthesizer), m_args(args)
{
}

// Translates a wxArrayString into an argc, argv pair.
// Since each method is called from one site, its all defined inline.
class Cmd_line
{
  int m_argc;
  const char **m_argv;
public:
  Cmd_line(const wxArrayString& args)
  {
    m_argc = args.GetCount();
    m_argv = new const char*[m_argc + 1];
    for (int i = 0; i < m_argc; i++) {
      const wxWX2MBbuf buf = args[i].mb_str(wxConvUTF8);
      const char *arg = (const char *)buf;
      size_t n = strlen(arg);
      m_argv[i] = strcpy(new char[n + 1], arg);
    }
    m_argv[m_argc] = 0;
  }
  ~Cmd_line()
  {
    for (int i = 0; i < m_argc; i++)
      delete [] m_argv[i];
    delete [] m_argv;
  }
  int get_argc() const
  {
    return m_argc;
  }
  char **get_argv() const
  {
    return const_cast<char **>(m_argv);
  }
};

// Why does this return void * rather than ExitCode?
void *Worker_thread::Entry()
{
  Cmd_line cmd_line(m_args);
  int argc = cmd_line.get_argc();
  char **argv = cmd_line.get_argv();
  int rc = m_synthesizer->perform(argc, argv);
  Main& app = wxGetApp();
  app.post_halt(rc);
  return (ExitCode)0;
}

// Called only in the worker thread
bool Worker_thread::poll_events()
{
  return !TestDestroy();
}
